from .darcy import load_darcy_pt, load_darcy_flow_small
from .spherical_swe import load_spherical_swe
from .navier_stokes import load_navier_stokes_pt 
from .pt_dataset import load_pt_traintestsplit
from .burgers import load_burgers_1dtime
